@@ -1,4 +1,7 @@
 /*
+ * $PostgreSQL: pgsql/contrib/tablefunc/tablefunc.c,v 1.60.2.1 2009/12/29 17:41:09 heikki Exp $
+ *
+ *
  * tablefunc
  *
  * Sample to demonstrate C functions which return setof scalar
@@ -7,7 +10,7 @@
  * And contributors:
  * Nabil Sayegh <postgresql@e-trolley.de>
  *
- * Copyright (c) 2002-2008, PostgreSQL Global Development Group
+ * Copyright (c) 2002-2009, PostgreSQL Global Development Group
  *
  * Permission to use, copy, modify, and distribute this software and its
  * documentation for any purpose, without fee, and without a written agreement
@@ -31,6 +34,7 @@
 
 #include <math.h>
 
+#include "catalog/pg_type.h"
 #include "fmgr.h"
 #include "funcapi.h"
 #include "executor/spi.h"
@@ -48,7 +52,8 @@ static HTAB *load_categories_hash(char *cats_sql, MemoryContext per_query_ctx);
 static Tuplestorestate *get_crosstab_tuplestore(char *sql,
 						HTAB *crosstab_hash,
 						TupleDesc tupdesc,
-						MemoryContext per_query_ctx);
+						MemoryContext per_query_ctx,
+						bool randomAccess);
 static void validateConnectbyTupleDesc(TupleDesc tupdesc, bool show_branch, bool show_serial);
 static bool compatCrosstabTupleDescs(TupleDesc tupdesc1, TupleDesc tupdesc2);
 static bool compatConnectbyTupleDescs(TupleDesc tupdesc1, TupleDesc tupdesc2);
@@ -63,6 +68,7 @@ static Tuplestorestate *connectby(char *relname,
 		  bool show_branch,
 		  bool show_serial,
 		  MemoryContext per_query_ctx,
+		  bool randomAccess,
 		  AttInMetadata *attinmeta);
 static Tuplestorestate *build_tuplestore_recursively(char *key_fld,
 							 char *parent_key_fld,
@@ -87,16 +93,8 @@ typedef struct
 	float8		stddev;			/* stddev of the distribution */
 	float8		carry_val;		/* hold second generated value */
 	bool		use_carry;		/* use second generated value */
-}	normal_rand_fctx;
+} normal_rand_fctx;
 
-typedef struct
-{
-	SPITupleTable *spi_tuptable;	/* sql results from user query */
-	char	   *lastrowid;		/* rowid of the last tuple sent */
-}	crosstab_fctx;
-
-#define GET_TEXT(cstrp) DatumGetTextP(DirectFunctionCall1(textin, CStringGetDatum(cstrp)))
-#define GET_STR(textp) DatumGetCString(DirectFunctionCall1(textout, PointerGetDatum(textp)))
 #define xpfree(var_) \
 	do { \
 		if (var_ != NULL) \
@@ -126,7 +124,7 @@ typedef struct crosstab_cat_desc
 {
 	char	   *catname;		/* full category name */
 	int			attidx;			/* zero based */
-}	crosstab_cat_desc;
+} crosstab_cat_desc;
 
 #define MAX_CATNAME_LEN			NAMEDATALEN
 #define INIT_CATS				64
@@ -165,7 +163,7 @@ typedef struct crosstab_hashent
 {
 	char		internal_catname[MAX_CATNAME_LEN];
 	crosstab_cat_desc *catdesc;
-}	crosstab_HashEnt;
+} crosstab_HashEnt;
 
 /*
  * normal_rand - return requested number of random values
@@ -353,304 +351,249 @@ PG_FUNCTION_INFO_V1(crosstab);
 Datum
 crosstab(PG_FUNCTION_ARGS)
 {
-	FuncCallContext *funcctx;
-	TupleDesc	ret_tupdesc;
+	char	   *sql = text_to_cstring(PG_GETARG_TEXT_PP(0));
+	ReturnSetInfo *rsinfo = (ReturnSetInfo *) fcinfo->resultinfo;
+	Tuplestorestate *tupstore;
+	TupleDesc	tupdesc;
 	int			call_cntr;
 	int			max_calls;
 	AttInMetadata *attinmeta;
-	SPITupleTable *spi_tuptable = NULL;
+	SPITupleTable *spi_tuptable;
 	TupleDesc	spi_tupdesc;
-	char	   *lastrowid = NULL;
-	crosstab_fctx *fctx;
+	bool		firstpass;
+	char	   *lastrowid;
 	int			i;
 	int			num_categories;
-	bool		firstpass = false;
+	MemoryContext per_query_ctx;
 	MemoryContext oldcontext;
+	int			ret;
+	int			proc;
 
-	/* stuff done only on the first call of the function */
-	if (SRF_IS_FIRSTCALL())
+	/* check to see if caller supports us returning a tuplestore */
+	if (rsinfo == NULL || !IsA(rsinfo, ReturnSetInfo))
+		ereport(ERROR,
+				(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
+				 errmsg("set-valued function called in context that cannot accept a set")));
+	if (!(rsinfo->allowedModes & SFRM_Materialize))
+		ereport(ERROR,
+				(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
+				 errmsg("materialize mode required, but it is not " \
+						"allowed in this context")));
+
+	per_query_ctx = rsinfo->econtext->ecxt_per_query_memory;
+
+	/* Connect to SPI manager */
+	if ((ret = SPI_connect()) < 0)
+		/* internal error */
+		elog(ERROR, "crosstab: SPI_connect returned %d", ret);
+
+	/* Retrieve the desired rows */
+	ret = SPI_execute(sql, true, 0);
+	proc = SPI_processed;
+
+	/* If no qualifying tuples, fall out early */
+	if (ret != SPI_OK_SELECT || proc <= 0)
 	{
-		char	   *sql = GET_STR(PG_GETARG_TEXT_P(0));
-		TupleDesc	tupdesc;
-		int			ret;
-		int			proc;
-
-		/* create a function context for cross-call persistence */
-		funcctx = SRF_FIRSTCALL_INIT();
-
-		/* Connect to SPI manager */
-		if ((ret = SPI_connect()) < 0)
-			/* internal error */
-			elog(ERROR, "crosstab: SPI_connect returned %d", ret);
-
-		/* Retrieve the desired rows */
-		ret = SPI_execute(sql, true, 0);
-		proc = SPI_processed;
-
-		/* Check for qualifying tuples */
-		if ((ret == SPI_OK_SELECT) && (proc > 0))
-		{
-			spi_tuptable = SPI_tuptable;
-			spi_tupdesc = spi_tuptable->tupdesc;
-
-			/*----------
-			 * The provided SQL query must always return three columns.
-			 *
-			 * 1. rowname
-			 *	the label or identifier for each row in the final result
-			 * 2. category
-			 *	the label or identifier for each column in the final result
-			 * 3. values
-			 *	the value for each column in the final result
-			 *----------
-			 */
-			if (spi_tupdesc->natts != 3)
-				ereport(ERROR,
-						(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
-						 errmsg("invalid source data SQL statement"),
-						 errdetail("The provided SQL must return 3 "
-								   "columns: rowid, category, and values.")));
-		}
-		else
-		{
-			/* no qualifying tuples */
-			SPI_finish();
-			SRF_RETURN_DONE(funcctx);
-		}
-
-		/* get a tuple descriptor for our result type */
-		switch (get_call_result_type(fcinfo, NULL, &tupdesc))
-		{
-			case TYPEFUNC_COMPOSITE:
-				/* success */
-				break;
-			case TYPEFUNC_RECORD:
-				/* failed to determine actual type of RECORD */
-				ereport(ERROR,
-						(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
-						 errmsg("function returning record called in context "
-								"that cannot accept type record")));
-				break;
-			default:
-				/* result type isn't composite */
-				elog(ERROR, "return type must be a row type");
-				break;
-		}
-
-		/*
-		 * Check that return tupdesc is compatible with the data we got from
-		 * SPI, at least based on number and type of attributes
-		 */
-		if (!compatCrosstabTupleDescs(tupdesc, spi_tupdesc))
-			ereport(ERROR,
-					(errcode(ERRCODE_SYNTAX_ERROR),
-					 errmsg("return and sql tuple descriptions are " \
-							"incompatible")));
-
-		/*
-		 * switch to memory context appropriate for multiple function calls
-		 */
-		oldcontext = MemoryContextSwitchTo(funcctx->multi_call_memory_ctx);
-
-		/* make sure we have a persistent copy of the tupdesc */
-		tupdesc = CreateTupleDescCopy(tupdesc);
-
-		/*
-		 * Generate attribute metadata needed later to produce tuples from raw
-		 * C strings
-		 */
-		attinmeta = TupleDescGetAttInMetadata(tupdesc);
-		funcctx->attinmeta = attinmeta;
-
-		/* allocate memory for user context */
-		fctx = (crosstab_fctx *) palloc(sizeof(crosstab_fctx));
-
-		/*
-		 * Save spi data for use across calls
-		 */
-		fctx->spi_tuptable = spi_tuptable;
-		fctx->lastrowid = NULL;
-		funcctx->user_fctx = fctx;
-
-		/* total number of tuples to be returned */
-		funcctx->max_calls = proc;
-
-		MemoryContextSwitchTo(oldcontext);
-		firstpass = true;
+		SPI_finish();
+		rsinfo->isDone = ExprEndResult;
+		PG_RETURN_NULL();
 	}
 
-	/* stuff done on every call of the function */
-	funcctx = SRF_PERCALL_SETUP();
-
-	/*
-	 * initialize per-call variables
-	 */
-	call_cntr = funcctx->call_cntr;
-	max_calls = funcctx->max_calls;
-
-	/* user context info */
-	fctx = (crosstab_fctx *) funcctx->user_fctx;
-	lastrowid = fctx->lastrowid;
-	spi_tuptable = fctx->spi_tuptable;
-
-	/* the sql tuple */
+	spi_tuptable = SPI_tuptable;
 	spi_tupdesc = spi_tuptable->tupdesc;
 
-	/* attribute return type and return tuple description */
-	attinmeta = funcctx->attinmeta;
-	ret_tupdesc = attinmeta->tupdesc;
+	/*----------
+	 * The provided SQL query must always return three columns.
+	 *
+	 * 1. rowname
+	 *	the label or identifier for each row in the final result
+	 * 2. category
+	 *	the label or identifier for each column in the final result
+	 * 3. values
+	 *	the value for each column in the final result
+	 *----------
+	 */
+	if (spi_tupdesc->natts != 3)
+		ereport(ERROR,
+				(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
+				 errmsg("invalid source data SQL statement"),
+				 errdetail("The provided SQL must return 3 "
+						   "columns: rowid, category, and values.")));
+
+	/* get a tuple descriptor for our result type */
+	switch (get_call_result_type(fcinfo, NULL, &tupdesc))
+	{
+		case TYPEFUNC_COMPOSITE:
+			/* success */
+			break;
+		case TYPEFUNC_RECORD:
+			/* failed to determine actual type of RECORD */
+			ereport(ERROR,
+					(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
+					 errmsg("function returning record called in context "
+							"that cannot accept type record")));
+			break;
+		default:
+			/* result type isn't composite */
+			elog(ERROR, "return type must be a row type");
+			break;
+	}
+
+	/*
+	 * Check that return tupdesc is compatible with the data we got from SPI,
+	 * at least based on number and type of attributes
+	 */
+	if (!compatCrosstabTupleDescs(tupdesc, spi_tupdesc))
+		ereport(ERROR,
+				(errcode(ERRCODE_SYNTAX_ERROR),
+				 errmsg("return and sql tuple descriptions are " \
+						"incompatible")));
+
+	/*
+	 * switch to long-lived memory context
+	 */
+	oldcontext = MemoryContextSwitchTo(per_query_ctx);
+
+	/* make sure we have a persistent copy of the result tupdesc */
+	tupdesc = CreateTupleDescCopy(tupdesc);
+
+	/* initialize our tuplestore in long-lived context */
+	tupstore =
+		tuplestore_begin_heap(rsinfo->allowedModes & SFRM_Materialize_Random,
+							  false, work_mem);
+
+	MemoryContextSwitchTo(oldcontext);
+
+	/*
+	 * Generate attribute metadata needed later to produce tuples from raw C
+	 * strings
+	 */
+	attinmeta = TupleDescGetAttInMetadata(tupdesc);
+
+	/* total number of tuples to be examined */
+	max_calls = proc;
 
 	/* the return tuple always must have 1 rowid + num_categories columns */
-	num_categories = ret_tupdesc->natts - 1;
+	num_categories = tupdesc->natts - 1;
 
-	if (call_cntr < max_calls)	/* do when there is more left to send */
+	firstpass = true;
+	lastrowid = NULL;
+
+	for (call_cntr = 0; call_cntr < max_calls; call_cntr++)
 	{
-		HeapTuple	tuple;
-		Datum		result;
-		char	  **values;
 		bool		skip_tuple = false;
+		char	  **values;
 
-		while (true)
+		/* allocate and zero space */
+		values = (char **) palloc0((1 + num_categories) * sizeof(char *));
+
+		/*
+		 * now loop through the sql results and assign each value in sequence
+		 * to the next category
+		 */
+		for (i = 0; i < num_categories; i++)
 		{
-			/* allocate space */
-			values = (char **) palloc((1 + num_categories) * sizeof(char *));
+			HeapTuple	spi_tuple;
+			char	   *rowid;
 
-			/* and make sure it's clear */
-			memset(values, '\0', (1 + num_categories) * sizeof(char *));
+			/* see if we've gone too far already */
+			if (call_cntr >= max_calls)
+				break;
+
+			/* get the next sql result tuple */
+			spi_tuple = spi_tuptable->vals[call_cntr];
+
+			/* get the rowid from the current sql result tuple */
+			rowid = SPI_getvalue(spi_tuple, spi_tupdesc, 1);
 
 			/*
-			 * now loop through the sql results and assign each value in
-			 * sequence to the next category
+			 * If this is the first pass through the values for this rowid,
+			 * set the first column to rowid
 			 */
-			for (i = 0; i < num_categories; i++)
+			if (i == 0)
 			{
-				HeapTuple	spi_tuple;
-				char	   *rowid = NULL;
-
-				/* see if we've gone too far already */
-				if (call_cntr >= max_calls)
-					break;
-
-				/* get the next sql result tuple */
-				spi_tuple = spi_tuptable->vals[call_cntr];
-
-				/* get the rowid from the current sql result tuple */
-				rowid = SPI_getvalue(spi_tuple, spi_tupdesc, 1);
+				xpstrdup(values[0], rowid);
 
 				/*
-				 * If this is the first pass through the values for this
-				 * rowid, set the first column to rowid
+				 * Check to see if the rowid is the same as that of the last
+				 * tuple sent -- if so, skip this tuple entirely
 				 */
-				if (i == 0)
+				if (!firstpass && xstreq(lastrowid, rowid))
 				{
-					xpstrdup(values[0], rowid);
-
-					/*
-					 * Check to see if the rowid is the same as that of the
-					 * last tuple sent -- if so, skip this tuple entirely
-					 */
-					if (!firstpass && xstreq(lastrowid, rowid))
-					{
-						skip_tuple = true;
-						break;
-					}
-				}
-
-				/*
-				 * If rowid hasn't changed on us, continue building the ouput
-				 * tuple.
-				 */
-				if (xstreq(rowid, values[0]))
-				{
-					/*
-					 * Get the next category item value, which is always
-					 * attribute number three.
-					 *
-					 * Be careful to assign the value to the array index based
-					 * on which category we are presently processing.
-					 */
-					values[1 + i] = SPI_getvalue(spi_tuple, spi_tupdesc, 3);
-
-					/*
-					 * increment the counter since we consume a row for each
-					 * category, but not for last pass because the API will do
-					 * that for us
-					 */
-					if (i < (num_categories - 1))
-						call_cntr = ++funcctx->call_cntr;
-				}
-				else
-				{
-					/*
-					 * We'll fill in NULLs for the missing values, but we need
-					 * to decrement the counter since this sql result row
-					 * doesn't belong to the current output tuple.
-					 */
-					call_cntr = --funcctx->call_cntr;
+					xpfree(rowid);
+					skip_tuple = true;
 					break;
 				}
-				xpfree(rowid);
 			}
 
 			/*
-			 * switch to memory context appropriate for multiple function
-			 * calls
+			 * If rowid hasn't changed on us, continue building the output
+			 * tuple.
 			 */
-			oldcontext = MemoryContextSwitchTo(funcctx->multi_call_memory_ctx);
-
-			xpfree(fctx->lastrowid);
-			xpstrdup(fctx->lastrowid, values[0]);
-			lastrowid = fctx->lastrowid;
-
-			MemoryContextSwitchTo(oldcontext);
-
-			if (!skip_tuple)
+			if (xstreq(rowid, values[0]))
 			{
-				/* build the tuple */
-				tuple = BuildTupleFromCStrings(attinmeta, values);
+				/*
+				 * Get the next category item value, which is always attribute
+				 * number three.
+				 *
+				 * Be careful to assign the value to the array index based on
+				 * which category we are presently processing.
+				 */
+				values[1 + i] = SPI_getvalue(spi_tuple, spi_tupdesc, 3);
 
-				/* make the tuple into a datum */
-				result = HeapTupleGetDatum(tuple);
-
-				/* Clean up */
-				for (i = 0; i < num_categories + 1; i++)
-					if (values[i] != NULL)
-						xpfree(values[i]);
-				xpfree(values);
-
-				SRF_RETURN_NEXT(funcctx, result);
+				/*
+				 * increment the counter since we consume a row for each
+				 * category, but not for last pass because the outer loop will
+				 * do that for us
+				 */
+				if (i < (num_categories - 1))
+					call_cntr++;
+				xpfree(rowid);
 			}
 			else
 			{
 				/*
-				 * Skipping this tuple entirely, but we need to advance the
-				 * counter like the API would if we had returned one.
+				 * We'll fill in NULLs for the missing values, but we need to
+				 * decrement the counter since this sql result row doesn't
+				 * belong to the current output tuple.
 				 */
-				call_cntr = ++funcctx->call_cntr;
-
-				/* we'll start over at the top */
-				xpfree(values);
-
-				/* see if we've gone too far already */
-				if (call_cntr >= max_calls)
-				{
-					/* release SPI related resources */
-					SPI_finish();
-					SRF_RETURN_DONE(funcctx);
-				}
-
-				/* need to reset this before the next tuple is started */
-				skip_tuple = false;
+				call_cntr--;
+				xpfree(rowid);
+				break;
 			}
 		}
+
+		if (!skip_tuple)
+		{
+			HeapTuple	tuple;
+
+			/* build the tuple and store it */
+			tuple = BuildTupleFromCStrings(attinmeta, values);
+			tuplestore_puttuple(tupstore, tuple);
+			heap_freetuple(tuple);
+		}
+
+		/* Remember current rowid */
+		xpfree(lastrowid);
+		xpstrdup(lastrowid, values[0]);
+		firstpass = false;
+
+		/* Clean up */
+		for (i = 0; i < num_categories + 1; i++)
+			if (values[i] != NULL)
+				pfree(values[i]);
+		pfree(values);
 	}
-	else
-		/* do when there is no more left */
-	{
-		/* release SPI related resources */
-		SPI_finish();
-		SRF_RETURN_DONE(funcctx);
-	}
+
+	/* let the caller know we're sending back a tuplestore */
+	rsinfo->returnMode = SFRM_Materialize;
+	rsinfo->setResult = tupstore;
+	rsinfo->setDesc = tupdesc;
+
+	/* release SPI related resources (and return to caller's context) */
+	SPI_finish();
+
+	return (Datum) 0;
 }
 
 /*
@@ -692,8 +635,8 @@ PG_FUNCTION_INFO_V1(crosstab_hash);
 Datum
 crosstab_hash(PG_FUNCTION_ARGS)
 {
-	char	   *sql = GET_STR(PG_GETARG_TEXT_P(0));
-	char	   *cats_sql = GET_STR(PG_GETARG_TEXT_P(1));
+	char	   *sql = text_to_cstring(PG_GETARG_TEXT_PP(0));
+	char	   *cats_sql = text_to_cstring(PG_GETARG_TEXT_PP(1));
 	ReturnSetInfo *rsinfo = (ReturnSetInfo *) fcinfo->resultinfo;
 	TupleDesc	tupdesc;
 	MemoryContext per_query_ctx;
@@ -705,7 +648,8 @@ crosstab_hash(PG_FUNCTION_ARGS)
 		ereport(ERROR,
 				(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
 				 errmsg("set-valued function called in context that cannot accept a set")));
-	if (!(rsinfo->allowedModes & SFRM_Materialize))
+	if (!(rsinfo->allowedModes & SFRM_Materialize) ||
+		rsinfo->expectedDesc == NULL)
 		ereport(ERROR,
 				(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
 				 errmsg("materialize mode required, but it is not " \
@@ -740,7 +684,8 @@ crosstab_hash(PG_FUNCTION_ARGS)
 	rsinfo->setResult = get_crosstab_tuplestore(sql,
 												crosstab_hash,
 												tupdesc,
-												per_query_ctx);
+												per_query_ctx,
+							 rsinfo->allowedModes & SFRM_Materialize_Random);
 
 	/*
 	 * SFRM_Materialize mode expects us to return a NULL Datum. The actual
@@ -847,7 +792,8 @@ static Tuplestorestate *
 get_crosstab_tuplestore(char *sql,
 						HTAB *crosstab_hash,
 						TupleDesc tupdesc,
-						MemoryContext per_query_ctx)
+						MemoryContext per_query_ctx,
+						bool randomAccess)
 {
 	Tuplestorestate *tupstore;
 	int			num_categories = hash_get_num_entries(crosstab_hash);
@@ -857,8 +803,8 @@ get_crosstab_tuplestore(char *sql,
 	int			ret;
 	int			proc;
 
-	/* initialize our tuplestore */
-	tupstore = tuplestore_begin_heap(true, false, work_mem);
+	/* initialize our tuplestore (while still in query context!) */
+	tupstore = tuplestore_begin_heap(randomAccess, false, work_mem);
 
 	/* Connect to SPI manager */
 	if ((ret = SPI_connect()) < 0)
@@ -1042,10 +988,10 @@ PG_FUNCTION_INFO_V1(connectby_text);
 Datum
 connectby_text(PG_FUNCTION_ARGS)
 {
-	char	   *relname = GET_STR(PG_GETARG_TEXT_P(0));
-	char	   *key_fld = GET_STR(PG_GETARG_TEXT_P(1));
-	char	   *parent_key_fld = GET_STR(PG_GETARG_TEXT_P(2));
-	char	   *start_with = GET_STR(PG_GETARG_TEXT_P(3));
+	char	   *relname = text_to_cstring(PG_GETARG_TEXT_PP(0));
+	char	   *key_fld = text_to_cstring(PG_GETARG_TEXT_PP(1));
+	char	   *parent_key_fld = text_to_cstring(PG_GETARG_TEXT_PP(2));
+	char	   *start_with = text_to_cstring(PG_GETARG_TEXT_PP(3));
 	int			max_depth = PG_GETARG_INT32(4);
 	char	   *branch_delim = NULL;
 	bool		show_branch = false;
@@ -1061,7 +1007,8 @@ connectby_text(PG_FUNCTION_ARGS)
 		ereport(ERROR,
 				(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
 				 errmsg("set-valued function called in context that cannot accept a set")));
-	if (!(rsinfo->allowedModes & SFRM_Materialize))
+	if (!(rsinfo->allowedModes & SFRM_Materialize) ||
+		rsinfo->expectedDesc == NULL)
 		ereport(ERROR,
 				(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
 				 errmsg("materialize mode required, but it is not " \
@@ -1069,7 +1016,7 @@ connectby_text(PG_FUNCTION_ARGS)
 
 	if (fcinfo->nargs == 6)
 	{
-		branch_delim = GET_STR(PG_GETARG_TEXT_P(5));
+		branch_delim = text_to_cstring(PG_GETARG_TEXT_PP(5));
 		show_branch = true;
 	}
 	else
@@ -1100,6 +1047,7 @@ connectby_text(PG_FUNCTION_ARGS)
 								  show_branch,
 								  show_serial,
 								  per_query_ctx,
+							  rsinfo->allowedModes & SFRM_Materialize_Random,
 								  attinmeta);
 	rsinfo->setDesc = tupdesc;
 
@@ -1119,16 +1067,15 @@ PG_FUNCTION_INFO_V1(connectby_text_serial);
 Datum
 connectby_text_serial(PG_FUNCTION_ARGS)
 {
-	char	   *relname = GET_STR(PG_GETARG_TEXT_P(0));
-	char	   *key_fld = GET_STR(PG_GETARG_TEXT_P(1));
-	char	   *parent_key_fld = GET_STR(PG_GETARG_TEXT_P(2));
-	char	   *orderby_fld = GET_STR(PG_GETARG_TEXT_P(3));
-	char	   *start_with = GET_STR(PG_GETARG_TEXT_P(4));
+	char	   *relname = text_to_cstring(PG_GETARG_TEXT_PP(0));
+	char	   *key_fld = text_to_cstring(PG_GETARG_TEXT_PP(1));
+	char	   *parent_key_fld = text_to_cstring(PG_GETARG_TEXT_PP(2));
+	char	   *orderby_fld = text_to_cstring(PG_GETARG_TEXT_PP(3));
+	char	   *start_with = text_to_cstring(PG_GETARG_TEXT_PP(4));
 	int			max_depth = PG_GETARG_INT32(5);
 	char	   *branch_delim = NULL;
 	bool		show_branch = false;
 	bool		show_serial = true;
-
 	ReturnSetInfo *rsinfo = (ReturnSetInfo *) fcinfo->resultinfo;
 	TupleDesc	tupdesc;
 	AttInMetadata *attinmeta;
@@ -1140,7 +1087,8 @@ connectby_text_serial(PG_FUNCTION_ARGS)
 		ereport(ERROR,
 				(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
 				 errmsg("set-valued function called in context that cannot accept a set")));
-	if (!(rsinfo->allowedModes & SFRM_Materialize))
+	if (!(rsinfo->allowedModes & SFRM_Materialize) ||
+		rsinfo->expectedDesc == NULL)
 		ereport(ERROR,
 				(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
 				 errmsg("materialize mode required, but it is not " \
@@ -1148,7 +1096,7 @@ connectby_text_serial(PG_FUNCTION_ARGS)
 
 	if (fcinfo->nargs == 7)
 	{
-		branch_delim = GET_STR(PG_GETARG_TEXT_P(6));
+		branch_delim = text_to_cstring(PG_GETARG_TEXT_PP(6));
 		show_branch = true;
 	}
 	else
@@ -1179,6 +1127,7 @@ connectby_text_serial(PG_FUNCTION_ARGS)
 								  show_branch,
 								  show_serial,
 								  per_query_ctx,
+							  rsinfo->allowedModes & SFRM_Materialize_Random,
 								  attinmeta);
 	rsinfo->setDesc = tupdesc;
 
@@ -1209,6 +1158,7 @@ connectby(char *relname,
 		  bool show_branch,
 		  bool show_serial,
 		  MemoryContext per_query_ctx,
+		  bool randomAccess,
 		  AttInMetadata *attinmeta)
 {
 	Tuplestorestate *tupstore = NULL;
@@ -1226,7 +1176,7 @@ connectby(char *relname,
 	oldcontext = MemoryContextSwitchTo(per_query_ctx);
 
 	/* initialize our tuplestore */
-	tupstore = tuplestore_begin_heap(true, false, work_mem);
+	tupstore = tuplestore_begin_heap(randomAccess, false, work_mem);
 
 	MemoryContextSwitchTo(oldcontext);
 
@@ -1583,6 +1533,10 @@ compatCrosstabTupleDescs(TupleDesc ret_tupdesc, TupleDesc sql_tupdesc)
 	Form_pg_attribute sql_attr;
 	Oid			sql_atttypid;
 
+	if (ret_tupdesc->natts < 2 ||
+		sql_tupdesc->natts < 3)
+		return false;
+
 	/* check the rowid types match */
 	ret_atttypid = ret_tupdesc->attrs[0]->atttypid;
 	sql_atttypid = sql_tupdesc->attrs[0]->atttypid;
@@ -1622,9 +1576,10 @@ quote_literal_cstr(char *rawstr)
 	text	   *result_text;
 	char	   *result;
 
-	rawstr_text = DatumGetTextP(DirectFunctionCall1(textin, CStringGetDatum(rawstr)));
-	result_text = DatumGetTextP(DirectFunctionCall1(quote_literal, PointerGetDatum(rawstr_text)));
-	result = DatumGetCString(DirectFunctionCall1(textout, PointerGetDatum(result_text)));
+	rawstr_text = cstring_to_text(rawstr);
+	result_text = DatumGetTextP(DirectFunctionCall1(quote_literal,
+											  PointerGetDatum(rawstr_text)));
+	result = text_to_cstring(result_text);
 
 	return result;
 }
